# coding: utf-8
from typing import List, Type

from django.core.exceptions import ValidationError
from django.db.models import Model
from m3 import ApplicationLogicException
from ninja import Query
from ninja import Router
from ninja.responses import codes_4xx
from ninja.responses import codes_5xx

from .helpers import create_error_response
from .helpers import create_search_schema
from .schema import DeletedObjSchema
from .schema import EnumSchema
from .schema import ErrorResponseSchema
from .schema import ListSchema
from .service import BaseServiceProvider


class ApiModel:
    """Базовый api для работы с моделями."""
    @property
    def response_schema(self):
        """Данные, возвращемые при создании/обновлении объекта модели."""
        raise NotImplementedError(
            "Необходимо указать схему данных для ответа"
        )

    @property
    def create_schema(self):
        """Данные, по которым будет создаваться объект модели."""
        raise NotImplementedError(
            "Необходимо указать схему данных для создания"
        )

    @property
    def update_schema(self):
        """Данные, по которым будет обновляться объект модели."""
        raise NotImplementedError(
            "Необходимо указать схему данных для обновления"
        )

    @property
    def router(self) -> Router:
        """Роутер для подключения к основному Api проекта."""
        if not getattr(self, "_router", None):
            self._router = Router()
        return self._router

    @property
    def provider_class(self) -> Type[BaseServiceProvider]:
        """Класс провайдера данных."""
        return BaseServiceProvider

    @property
    def provider(self) -> BaseServiceProvider:
        """Провайдер для работы с данными."""
        return self._provider

    @property
    def model(self):
        """Модель для доступа к данным."""
        raise NotImplementedError(
            "Необходимо указать модель"
        )

    # Перечень полей модели для поиска
    search_fields = None

    def get_list_method_summary(self):
        return f"Отображение объектов модели {self.model._meta.verbose_name}"

    def get_list_by_ids_method_summary(self):
        return (
            f"Отображение объектов модели {self.model._meta.verbose_name} "
            f"по списку id")

    def get_create_method_summary(self):
        return f"Создание объекта модели {self.model._meta.verbose_name}"

    def get_delete_method_summary(self):
        return f"Удаление объекта модели {self.model._meta.verbose_name}"

    def get_put_method_summary(self):
        return (f"Полное обновление объекта"
                f" модели {self.model._meta.verbose_name}")

    def get_putch_method_summary(self):
        return (f"Частичное обновление объекта"
                f" модели {self.model._meta.verbose_name}")

    def get_retrieved_method_summary(self):
        return f"Возвращение объекта модели {self.model._meta.verbose_name}"

    def __init__(self):
        self._init_provider(self.model)

        tags = [f"{self.model._meta.verbose_name_plural}"]
        self._add_main_api_operations(tags)
        self._add_additional_api_operations(tags)

    def _init_provider(self, model: Model):
        """Инициализирует провайдер данных."""
        self._provider = self.provider_class()
        self._provider.init(model=model)

    def _add_main_api_operations(self, tags):
        """Добавляет основные обработчики запросов"""
        self.router.add_api_operation(
            path="/", methods=["GET"],
            summary=self.get_list_method_summary(),
            view_func=self.get_list_func(),
            response={
                200: List[self.response_schema],
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/get_by_ids", methods=["GET"],
            summary=self.get_list_by_ids_method_summary(),
            view_func=self.get_list_finc_by_ids(),
            response={
                200: List[self.response_schema],
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/{pk}", methods=["GET"],
            summary=self.get_retrieved_method_summary(),
            view_func=self.get_retrieved_obj_func(),
            response={
                200: self.response_schema,
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/{pk}", methods=["PATCH"],
            summary=self.get_putch_method_summary(),
            view_func=self.get_patch_obj_func(),
            response={
                200: self.response_schema,
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/{pk}", methods=["PUT"],
            summary=self.get_put_method_summary(),
            view_func=self.get_update_func(),
            response={
                200: self.response_schema,
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/", methods=["POST"],
            summary=self.get_create_method_summary(),
            view_func=self.get_create_obj_func(),
            response={
                200: self.response_schema,
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )
        self.router.add_api_operation(
            path="/{pk}", methods=["DELETE"],
            summary=self.get_delete_method_summary(),
            view_func=self.get_delete_obj_func(),
            response={
                200: DeletedObjSchema,
                codes_4xx: ErrorResponseSchema,
                codes_5xx: ErrorResponseSchema,
            },
            tags=tags,
        )

        self._add_additional_api_operations(tags)

    def _add_additional_api_operations(self, tags):
        """Добавляет дополнительные обработчики запросов."""

    def get_list_schema(self):
        if self.search_fields:
            assert isinstance(self.search_fields, list)
            model_name = f'List{self.model._meta.object_name}'
            return create_search_schema(model_name, ListSchema,
                                        model=self.model,
                                        fields=self.search_fields
                                        )
        return ListSchema

    def get_list_func(self):
        """Возвращает функцию, возвращающую список объектов модели."""
        provider = self.provider
        search_fields = None
        if self.search_fields:
            assert isinstance(self.search_fields, list)
            search_fields = self.search_fields
        list_schema = self.get_list_schema()

        def list_func(request, filters: list_schema = Query(...)):
            """Возвращает список объектов модели."""
            params = filters.dict()
            limit = params.get('limit')
            offset = params.get('offset')
            ids = params.get('ids')
            f = params.get('f')
            order_by = params.get('order_by')

            if search_fields:
                search_fields_data = {
                    k: v for k, v in params.items()
                    if (k in search_fields and v is not None)
                }
                try:
                    response = provider.get_objects_by_search_fields(
                        limit, offset, search_fields_data, order_by
                    )
                except Exception as exc:
                    return create_error_response(500, exc)
            else:
                try:
                    response = provider.get_objects(
                        limit, offset, ids, f, order_by
                    )
                except Exception as exc:
                    return create_error_response(500, exc)

            return 200, response

        list_func.__name__ = f"list_{self.model._meta.model_name}"
        return list_func

    def get_retrieved_obj_func(self):
        """Возвращает функцию, возвращающую объект модели по pk."""
        provider = self.provider

        def retrieved_obj_func(request, pk: int):
            """Возвращает объект модели по pk."""
            try:
                response = provider.get_obj_by_id(pk)
            except Exception as exc:
                return create_error_response(500, exc)

            if not response:
                return create_error_response(
                    404,
                    Exception(f"Объект с id {pk} не найден.")
                )

            return 200, response

        retrieved_obj_func.__name__ = f"retrived_{self.model._meta.model_name}"
        return retrieved_obj_func

    def get_update_func(self):
        """Возвращает функцию, обновляющую все поля объект модели."""
        provider = self.provider
        update_schema = self.update_schema

        def update_func(request, pk: int, obj_data: update_schema):
            """Обновление объекта модели по заданным в schema данным."""
            try:
                response = provider.update_obj(pk, obj_data.dict())
            except Exception as exc:
                return create_error_response(500, exc)

            return 200, response

        update_func.__name__ = f"update_{self.model._meta.model_name}"
        return update_func

    def get_create_obj_func(self):
        """Возвращает функцию, создающую объект модели."""
        provider = self.provider
        create_schema = self.create_schema

        def create_func(request, obj_data: create_schema):
            """Создание объекта модели по заданным в schema данным."""
            try:
                response = provider.create_obj(obj_data.dict())
            except ValidationError as exc:
                return create_error_response(400, exc)
            except Exception as exc:
                return create_error_response(500, exc)

            return 200, response

        create_func.__name__ = f"create_{self.model._meta.model_name}"
        return create_func

    def get_delete_obj_func(self):
        """Возвращает функцию, удаляющую объект модели по pk."""
        provider = self.provider

        def delete_func(request, pk: int):
            """
            Удаление объекта модели. Объект определяется по pk.
            """
            try:
                response = provider.delete_obj(pk)
            except ApplicationLogicException as exc:
                return create_error_response(400, exc)
            except Exception as exc:
                return create_error_response(500, exc)

            return 200, {"id": pk}

        delete_func.__name__ = f"delete_{self.model._meta.model_name}"
        return delete_func

    def get_patch_obj_func(self):
        """Возвращает функцию, частично обновляющую поля модели"""
        provider = self.provider
        update_schema = self.update_schema

        def patch_func(request, pk: int, obj_data: update_schema):
            """
            Частичное обновление объекта модели.
            Объект определяется по pk, поля для обновления в schema.
            """
            try:
                response = provider.update_obj(pk, obj_data.dict())
            except (ValidationError, ApplicationLogicException) as exc:
                return create_error_response(400, exc)
            except Exception as exc:
                return create_error_response(500, exc)

            return 200, response

        patch_func.__name__ = f"patch_{self.model._meta.model_name}"
        return patch_func

    def get_list_finc_by_ids(self):
        """Возвращает функцию, возвращающую список объектов модели по pk."""
        provider = self.provider

        def list_func(request, filters: EnumSchema = Query(...)):
            """Возвращает список объектов модели."""
            params = filters.dict()
            ids = params.get('ids', [])
            try:
                response = provider.get_objects_by_ids(ids)
            except Exception as exc:
                return create_error_response(500, exc)

            return 200, response

        list_func.__name__ = f"list_by_ids_{self.model._meta.model_name}"
        return list_func
